import pyomo.environ as pe
from pyomo.common.collections import ComponentMap
import pyomo.core.expr.numeric_expr as numeric_expr
from pyomo.core.expr.visitor import ExpressionValueVisitor
from pyomo.core.expr.numvalue import (
    nonpyomo_leaf_types, value, NumericValue, is_fixed, polynomial_degree, is_constant,
    native_numeric_types
)
from pyomo.core.expr.numeric_expr import ExpressionBase
from pyomo.contrib.fbbt.fbbt import compute_bounds_on_expr, fbbt
import math
from pyomo.core.base.constraint import Constraint
import logging
from .relaxations_base import BaseRelaxationData
from .univariate import PWUnivariateRelaxation, PWXSquaredRelaxation, PWCosRelaxation, PWSinRelaxation, PWArctanRelaxation
from .mccormick import PWMcCormickRelaxation
from .multivariate import MultivariateRelaxation
from .alphabb import AlphaBBRelaxation
from coramin.utils.coramin_enums import RelaxationSide, FunctionShape, Effort, EigenValueBounder
from pyomo.gdp import Disjunct
from pyomo.core.base.expression import _GeneralExpressionData, SimpleExpression
from coramin.relaxations.iterators import nonrelaxation_component_data_objects
from pyomo.contrib import appsi
from pyomo.repn.standard_repn import generate_standard_repn
from pyomo.contrib.fbbt import interval
from pyomo.core.expr.compare import convert_expression_to_prefix_notation
from .split_expr import split_expr
from coramin.utils.pyomo_utils import simplify_expr
from .hessian import Hessian
from typing import MutableMapping, Tuple, Union, Optional
from pyomo.core.base.block import _BlockData
from .iterators import relaxation_data_objects


logger = logging.getLogger(__name__)


class Hashable:
    def __init__(self, *args):
        entries = list()
        for i in args:
            itype = type(i)
            if itype is tuple or itype in nonpyomo_leaf_types:
                entries.append(i)
            elif isinstance(i, NumericValue):
                entries.append(id(i))
            else:
                raise NotImplementedError(
                    f'unexpected entry: {str(i)}')
        self.entries = entries
        self.hashable_entries = tuple(entries)

    def __eq__(self, other):
        if isinstance(other, Hashable):
            return self.entries == other.entries
        return False

    def __hash__(self):
        return hash(self.hashable_entries)


class RelaxationException(Exception):
    pass


class RelaxationCounter(object):
    def __init__(self):
        self.count = 0

    def increment(self):
        self.count += 1

    def __str__(self):
        return str(self.count)


def compute_float_bounds_on_expr(expr):
    lb, ub = compute_bounds_on_expr(expr)
    if lb is None:
        lb = -math.inf
    if ub is None:
        ub = math.inf

    return lb, ub


def replace_sub_expression_with_aux_var(arg, parent_block):
    if type(arg) in nonpyomo_leaf_types:
        return arg
    elif arg.is_expression_type():
        _var = parent_block.aux_vars.add()
        _con = parent_block.aux_cons.add(_var == arg)
        fbbt(_con)
        return _var
    else:
        return arg


def _get_aux_var(parent_block, expr):
    _aux_var = parent_block.aux_vars.add()
    lb, ub = compute_bounds_on_expr(expr)
    _aux_var.setlb(lb)
    _aux_var.setub(ub)
    try:
        expr_value = pe.value(expr, exception=False)
    except ArithmeticError:
        expr_value = None
    if expr_value is not None and pe.value(_aux_var, exception=False) is None:
        _aux_var.set_value(expr_value, skip_validation=True)
    return _aux_var


def _relax_leaf_to_root_ProductExpression(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg1, arg2 = values

    # The purpose of the next bit of code is to find common quadratic terms. For example, suppose we are relaxing
    # a model with the following two constraints:
    #
    # w1 - x*y = 0
    # w2 + 3*x*y = 0
    #
    # we want to end up with
    #
    # w1 - aux1 = 0
    # w2 + 3*aux1 = 0
    # aux1 = x*y
    #
    # rather than
    #
    # w1 - aux1 = 0
    # w2 + 3*aux2 = 0
    # aux1 = x*y
    # aux2 = x*y
    #

    h1 = Hashable(arg1, arg2, 'mul')
    h2 = Hashable(arg2, arg1, 'mul')
    if h1 in aux_var_map or h2 in aux_var_map:
        if h1 in aux_var_map:
            _aux_var, relaxation = aux_var_map[h1]
        else:
            _aux_var, relaxation = aux_var_map[h2]
        relaxation_side = relaxation_side_map[node]
        if relaxation_side != relaxation.relaxation_side:
            relaxation.relaxation_side = RelaxationSide.BOTH
        res = _aux_var
        degree_map[res] = 1
    else:
        degree_1 = degree_map[arg1]
        degree_2 = degree_map[arg2]
        if degree_1 == 0:
            res = arg1 * arg2
            degree_map[res] = degree_2
        elif degree_2 == 0:
            res = arg2 * arg1
            degree_map[res] = degree_1
        elif arg1.__class__ == numeric_expr.MonomialTermExpression or arg2.__class__ == numeric_expr.MonomialTermExpression:
            if arg1.__class__ == numeric_expr.MonomialTermExpression:
                coef1, arg1 = arg1.args
            else:
                coef1 = 1
            if arg2.__class__ == numeric_expr.MonomialTermExpression:
                coef2, arg2 = arg2.args
            else:
                coef2 = 1
            coef = coef1 * coef2
            _new_relaxation_side_map = ComponentMap()
            _reformulated = coef * (arg1 * arg2)
            _new_relaxation_side_map[_reformulated] = relaxation_side_map[node]
            res = _relax_expr(expr=_reformulated, aux_var_map=aux_var_map, parent_block=parent_block,
                              relaxation_side_map=_new_relaxation_side_map, counter=counter, degree_map=degree_map)
            degree_map[res] = 1
        elif arg1 is arg2:
            # reformulate arg1 * arg2 as arg1**2
            _new_relaxation_side_map = ComponentMap()
            _reformulated = arg1**2
            _new_relaxation_side_map[_reformulated] = relaxation_side_map[node]
            res = _relax_expr(expr=_reformulated, aux_var_map=aux_var_map, parent_block=parent_block,
                              relaxation_side_map=_new_relaxation_side_map, counter=counter, degree_map=degree_map)
            degree_map[res] = 1
        else:
            _aux_var = _get_aux_var(parent_block, arg1 * arg2)
            arg1 = replace_sub_expression_with_aux_var(arg1, parent_block)
            arg2 = replace_sub_expression_with_aux_var(arg2, parent_block)
            relaxation_side = relaxation_side_map[node]
            relaxation = PWMcCormickRelaxation()
            relaxation.set_input(x1=arg1, x2=arg2, aux_var=_aux_var, relaxation_side=relaxation_side)
            aux_var_map[h1] = (_aux_var, relaxation)
            setattr(parent_block.relaxations, 'rel'+str(counter), relaxation)
            counter.increment()
            res = _aux_var
            degree_map[res] = 1
    return res


def _relax_leaf_to_root_DivisionExpression(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg1, arg2 = values
    h1 = Hashable(arg1, arg2, 'div')
    if arg1.__class__ == numeric_expr.MonomialTermExpression:
        coef1, arg1 = arg1.args
    else:
        coef1 = 1
    if arg2.__class__ == numeric_expr.MonomialTermExpression:
        coef2, arg2 = arg2.args
    else:
        coef2 = 1
    coef = coef1/coef2
    degree_1 = degree_map[arg1]
    degree_2 = degree_map[arg2]

    if degree_2 == 0:
        res = (coef / arg2) * arg1
        degree_map[res] = degree_1
        return res
    elif h1 in aux_var_map:
        _aux_var, relaxation = aux_var_map[h1]
        relaxation_side = relaxation_side_map[node]
        if relaxation_side != relaxation.relaxation_side:
            relaxation.relaxation_side = RelaxationSide.BOTH
        res = coef * _aux_var
        degree_map[_aux_var] = 1
        degree_map[res] = 1
        return res
    elif degree_1 == 0:
        h2 = Hashable(arg2, 'reciprocal')
        if h2 in aux_var_map:
            _aux_var, relaxation = aux_var_map[h2]
            relaxation_side = relaxation_side_map[node]
            if relaxation_side != relaxation.relaxation_side:
                relaxation.relaxation_side = RelaxationSide.BOTH
            res = coef * arg1 * _aux_var
            degree_map[_aux_var] = 1
            degree_map[res] = 1
            return res
        else:
            _aux_var = _get_aux_var(parent_block, 1/arg2)
            arg2 = replace_sub_expression_with_aux_var(arg2, parent_block)
            relaxation_side = relaxation_side_map[node]
            degree_map[_aux_var] = 1
            if compute_float_bounds_on_expr(arg2)[0] > 0:
                relaxation = PWUnivariateRelaxation()
                relaxation.set_input(x=arg2, aux_var=_aux_var, relaxation_side=relaxation_side, f_x_expr=1/arg2,
                                     shape=FunctionShape.CONVEX)
            elif compute_float_bounds_on_expr(arg2)[1] < 0:
                relaxation = PWUnivariateRelaxation()
                relaxation.set_input(x=arg2, aux_var=_aux_var, relaxation_side=relaxation_side, f_x_expr=1/arg2,
                                     shape=FunctionShape.CONCAVE)
            else:
                _one = parent_block.aux_vars.add()
                _one.fix(1)
                relaxation = PWMcCormickRelaxation()
                relaxation.set_input(x1=arg2, x2=_aux_var, aux_var=_one, relaxation_side=relaxation_side)
            aux_var_map[h2] = (_aux_var, relaxation)
            setattr(parent_block.relaxations, 'rel'+str(counter), relaxation)
            counter.increment()
            res = coef * arg1 * _aux_var
            degree_map[res] = 1
            return res
    else:
        _aux_var = _get_aux_var(parent_block, arg1 / arg2)
        arg1 = replace_sub_expression_with_aux_var(arg1, parent_block)
        arg2 = replace_sub_expression_with_aux_var(arg2, parent_block)
        relaxation_side = relaxation_side_map[node]
        arg2_lb, arg2_ub = compute_float_bounds_on_expr(arg2)
        if arg2_lb >= 0:
            if relaxation_side == RelaxationSide.UNDER:
                relaxation_side = RelaxationSide.OVER
            elif relaxation_side == RelaxationSide.OVER:
                relaxation_side = RelaxationSide.UNDER
            else:
                assert relaxation_side == RelaxationSide.BOTH
        elif arg2_ub <= 0:
            pass
        else:
            relaxation_side = RelaxationSide.BOTH
        relaxation = PWMcCormickRelaxation()
        relaxation.set_input(x1=arg2, x2=_aux_var, aux_var=arg1, relaxation_side=relaxation_side)
        aux_var_map[h1] = (_aux_var, relaxation)
        setattr(parent_block.relaxations, 'rel'+str(counter), relaxation)
        counter.increment()
        res = coef * _aux_var
        degree_map[_aux_var] = 1
        degree_map[res] = 1
        return res


def _relax_quadratic(arg1, aux_var_map, relaxation_side, degree_map, parent_block, counter):
    _aux_var = _get_aux_var(parent_block, arg1**2)
    arg1 = replace_sub_expression_with_aux_var(arg1, parent_block)
    degree_map[_aux_var] = 1
    relaxation = PWXSquaredRelaxation()
    relaxation.set_input(x=arg1, aux_var=_aux_var, relaxation_side=relaxation_side)
    aux_var_map[Hashable(arg1, 2, 'pow')] = (_aux_var, relaxation)
    setattr(parent_block.relaxations, 'rel' + str(counter), relaxation)
    counter.increment()
    return _aux_var


def _relax_convex_pow(arg1, arg2, aux_var_map, relaxation_side, degree_map, parent_block, counter, swap=False):
    _aux_var = _get_aux_var(parent_block, arg1**arg2)
    if swap:
        arg2 = replace_sub_expression_with_aux_var(arg2, parent_block)
        _x = arg2
    else:
        arg1 = replace_sub_expression_with_aux_var(arg1, parent_block)
        _x = arg1
    degree_map[_aux_var] = 1
    relaxation = PWUnivariateRelaxation()
    relaxation.set_input(x=_x, aux_var=_aux_var, relaxation_side=relaxation_side, f_x_expr=arg1 ** arg2,
                         shape=FunctionShape.CONVEX)
    aux_var_map[Hashable(arg1, arg2, 'pow')] = (_aux_var, relaxation)
    setattr(parent_block.relaxations, 'rel' + str(counter), relaxation)
    counter.increment()
    return _aux_var


def _relax_concave_pow(arg1, arg2, aux_var_map, relaxation_side, degree_map, parent_block, counter):
    _aux_var = _get_aux_var(parent_block, arg1 ** arg2)
    arg1 = replace_sub_expression_with_aux_var(arg1, parent_block)
    degree_map[_aux_var] = 1
    relaxation = PWUnivariateRelaxation()
    relaxation.set_input(x=arg1, aux_var=_aux_var, relaxation_side=relaxation_side, f_x_expr=arg1 ** arg2,
                         shape=FunctionShape.CONCAVE)
    aux_var_map[Hashable(arg1, arg2, 'pow')] = (_aux_var, relaxation)
    setattr(parent_block.relaxations, 'rel' + str(counter), relaxation)
    counter.increment()
    return _aux_var


def _relax_leaf_to_root_PowExpression(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg1, arg2 = values
    h = Hashable(arg1, arg2, 'pow')
    if h in aux_var_map:
        _aux_var, relaxation = aux_var_map[h]
        if relaxation_side_map[node] != relaxation.relaxation_side:
            relaxation.relaxation_side = RelaxationSide.BOTH
        degree_map[_aux_var] = 1
        return _aux_var
    else:
        degree1 = degree_map[arg1]
        degree2 = degree_map[arg2]
        if degree2 == 0:
            if degree1 == 0:
                res = arg1 ** arg2
                degree_map[res] = 0
                return res
            if not is_constant(arg2):
                logger.warning('Only constant exponents are supported: ' + str(arg1**arg2) + '\nReplacing ' + str(arg2) + ' with its value.')
            arg2 = pe.value(arg2)
            if arg2 == 1:
                return arg1
            elif arg2 == 0:
                res = 1
                degree_map[res] = 0
                return res
            elif arg2 == 2:
                return _relax_quadratic(arg1=arg1, aux_var_map=aux_var_map, relaxation_side=relaxation_side_map[node],
                                        degree_map=degree_map, parent_block=parent_block, counter=counter)
            elif arg2 >= 0:
                if arg2 == round(arg2):
                    if arg2 % 2 == 0 or compute_float_bounds_on_expr(arg1)[0] >= 0:
                        return _relax_convex_pow(arg1=arg1, arg2=arg2, aux_var_map=aux_var_map,
                                                 relaxation_side=relaxation_side_map[node], degree_map=degree_map,
                                                 parent_block=parent_block, counter=counter)
                    elif compute_float_bounds_on_expr(arg1)[1] <= 0:
                        return _relax_concave_pow(arg1=arg1, arg2=arg2, aux_var_map=aux_var_map,
                                                  relaxation_side=relaxation_side_map[node], degree_map=degree_map,
                                                  parent_block=parent_block, counter=counter)
                    else:  # reformulate arg1 ** arg2 as arg1 * arg1 ** (arg2 - 1)
                        _new_relaxation_side_map = ComponentMap()
                        _reformulated = arg1 * arg1 ** (arg2 - 1)
                        _new_relaxation_side_map[_reformulated] = relaxation_side_map[node]
                        res = _relax_expr(expr=_reformulated, aux_var_map=aux_var_map, parent_block=parent_block,
                                          relaxation_side_map=_new_relaxation_side_map, counter=counter,
                                          degree_map=degree_map)
                        degree_map[res] = 1
                        return res
                else:
                    if arg2 < 1:
                        return _relax_concave_pow(arg1=arg1, arg2=arg2, aux_var_map=aux_var_map,
                                                  relaxation_side=relaxation_side_map[node], degree_map=degree_map,
                                                  parent_block=parent_block, counter=counter)
                    else:
                        return _relax_convex_pow(arg1=arg1, arg2=arg2, aux_var_map=aux_var_map,
                                                 relaxation_side=relaxation_side_map[node], degree_map=degree_map,
                                                 parent_block=parent_block, counter=counter)
            else:
                if arg2 == round(arg2):
                    if compute_float_bounds_on_expr(arg1)[0] >= 0:
                        return _relax_convex_pow(arg1=arg1, arg2=arg2, aux_var_map=aux_var_map,
                                                 relaxation_side=relaxation_side_map[node], degree_map=degree_map,
                                                 parent_block=parent_block, counter=counter)
                    elif compute_float_bounds_on_expr(arg1)[1] <= 0:
                        if arg2 % 2 == 0:
                            return _relax_convex_pow(arg1=arg1, arg2=arg2, aux_var_map=aux_var_map,
                                                     relaxation_side=relaxation_side_map[node], degree_map=degree_map,
                                                     parent_block=parent_block, counter=counter)
                        else:
                            return _relax_concave_pow(arg1=arg1, arg2=arg2, aux_var_map=aux_var_map,
                                                      relaxation_side=relaxation_side_map[node], degree_map=degree_map,
                                                      parent_block=parent_block, counter=counter)
                    else:
                        # reformulate arg1 ** arg2 as 1 / arg1 ** (-arg2)
                        _new_relaxation_side_map = ComponentMap()
                        _reformulated = 1 / (arg1 ** (-arg2))
                        _new_relaxation_side_map[_reformulated] = relaxation_side_map[node]
                        res = _relax_expr(expr=_reformulated, aux_var_map=aux_var_map, parent_block=parent_block,
                                          relaxation_side_map=_new_relaxation_side_map, counter=counter,
                                          degree_map=degree_map)
                        degree_map[res] = 1
                        return res
                else:
                    assert compute_float_bounds_on_expr(arg1)[0] >= 0
                    return _relax_convex_pow(arg1=arg1, arg2=arg2, aux_var_map=aux_var_map,
                                             relaxation_side=relaxation_side_map[node], degree_map=degree_map,
                                             parent_block=parent_block, counter=counter)
        elif degree1 == 0:
            if not is_constant(arg1):
                logger.warning('Found {0} raised to a variable power. However, {0} does not appear to be constant (maybe '
                               'it is or depends on a mutable Param?). Replacing {0} with its value.'.format(str(arg1)))
                arg1 = pe.value(arg1)
            if arg1 < 0:
                raise ValueError('Cannot raise a negative base to a variable exponent: ' + str(arg1**arg2))
            return _relax_convex_pow(arg1=arg1, arg2=arg2, aux_var_map=aux_var_map,
                                     relaxation_side=relaxation_side_map[node], degree_map=degree_map,
                                     parent_block=parent_block, counter=counter, swap=True)
        else:
            assert compute_float_bounds_on_expr(arg1)[0] >= 0
            _new_relaxation_side_map = ComponentMap()
            _reformulated = pe.exp(arg2 * pe.log(arg1))
            _new_relaxation_side_map[_reformulated] = relaxation_side_map[node]
            res = _relax_expr(expr=_reformulated, aux_var_map=aux_var_map, parent_block=parent_block,
                              relaxation_side_map=_new_relaxation_side_map, counter=counter,
                              degree_map=degree_map)
            degree_map[res] = 1
            return res


def _relax_leaf_to_root_SumExpression(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    res = sum(values)
    degree_map[res] = max([degree_map[arg] for arg in values])
    return res


def _relax_leaf_to_root_NegationExpression(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg = values[0]
    res = -arg
    degree_map[res] = degree_map[arg]
    return res


def _relax_leaf_to_root_sqrt(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg = values[0]
    _new_relaxation_side_map = ComponentMap()
    _reformulated = arg**0.5
    _new_relaxation_side_map[_reformulated] = relaxation_side_map[node]
    res = _relax_expr(expr=_reformulated, aux_var_map=aux_var_map, parent_block=parent_block,
                      relaxation_side_map=_new_relaxation_side_map, counter=counter,
                      degree_map=degree_map)
    degree_map[res] = 1
    return res


def _relax_leaf_to_root_exp(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg = values[0]
    degree = degree_map[arg]
    if degree == 0:
        res = pe.exp(arg)
        degree_map[res] = 0
        return res
    elif (id(arg), 'exp') in aux_var_map:
        _aux_var, relaxation = aux_var_map[id(arg), 'exp']
        relaxation_side = relaxation_side_map[node]
        if relaxation_side != relaxation.relaxation_side:
            relaxation.relaxation_side = RelaxationSide.BOTH
        degree_map[_aux_var] = 1
        return _aux_var
    else:
        _aux_var = _get_aux_var(parent_block, pe.exp(arg))
        arg = replace_sub_expression_with_aux_var(arg, parent_block)
        relaxation_side = relaxation_side_map[node]
        degree_map[_aux_var] = 1
        relaxation = PWUnivariateRelaxation()
        relaxation.set_input(x=arg, aux_var=_aux_var, relaxation_side=relaxation_side, f_x_expr=pe.exp(arg),
                             shape=FunctionShape.CONVEX)
        aux_var_map[id(arg), 'exp'] = (_aux_var, relaxation)
        setattr(parent_block.relaxations, 'rel'+str(counter), relaxation)
        counter.increment()
        return _aux_var


def _relax_leaf_to_root_log(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg = values[0]
    degree = degree_map[arg]
    if degree == 0:
        res = pe.exp(arg)
        degree_map[res] = 0
        return res
    elif (id(arg), 'log') in aux_var_map:
        _aux_var, relaxation = aux_var_map[id(arg), 'log']
        relaxation_side = relaxation_side_map[node]
        if relaxation_side != relaxation.relaxation_side:
            relaxation.relaxation_side = RelaxationSide.BOTH
        degree_map[_aux_var] = 1
        return _aux_var
    else:
        _aux_var = _get_aux_var(parent_block, pe.log(arg))
        arg = replace_sub_expression_with_aux_var(arg, parent_block)
        relaxation_side = relaxation_side_map[node]
        degree_map[_aux_var] = 1
        relaxation = PWUnivariateRelaxation()
        relaxation.set_input(x=arg, aux_var=_aux_var, relaxation_side=relaxation_side, f_x_expr=pe.log(arg),
                             shape=FunctionShape.CONCAVE)
        aux_var_map[id(arg), 'log'] = (_aux_var, relaxation)
        setattr(parent_block.relaxations, 'rel'+str(counter), relaxation)
        counter.increment()
        return _aux_var


def _relax_leaf_to_root_log10(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg = values[0]
    degree = degree_map[arg]
    if degree == 0:
        res = pe.exp(arg)
        degree_map[res] = 0
        return res
    elif (id(arg), 'log10') in aux_var_map:
        _aux_var, relaxation = aux_var_map[id(arg), 'log10']
        relaxation_side = relaxation_side_map[node]
        if relaxation_side != relaxation.relaxation_side:
            relaxation.relaxation_side = RelaxationSide.BOTH
        degree_map[_aux_var] = 1
        return _aux_var
    else:
        _aux_var = _get_aux_var(parent_block, pe.log10(arg))
        arg = replace_sub_expression_with_aux_var(arg, parent_block)
        relaxation_side = relaxation_side_map[node]
        degree_map[_aux_var] = 1
        relaxation = PWUnivariateRelaxation()
        relaxation.set_input(x=arg, aux_var=_aux_var, relaxation_side=relaxation_side, f_x_expr=pe.log10(arg),
                             shape=FunctionShape.CONCAVE)
        aux_var_map[id(arg), 'log10'] = (_aux_var, relaxation)
        setattr(parent_block.relaxations, 'rel'+str(counter), relaxation)
        counter.increment()
        return _aux_var


def _relax_leaf_to_root_sin(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg = values[0]
    degree = degree_map[arg]
    if degree == 0:
        res = pe.sin(arg)
        degree_map[res] = 0
        return res
    elif (id(arg), 'sin') in aux_var_map:
        _aux_var, relaxation = aux_var_map[id(arg), 'sin']
        relaxation_side = relaxation_side_map[node]
        if relaxation_side != relaxation.relaxation_side:
            relaxation.relaxation_side = RelaxationSide.BOTH
        degree_map[_aux_var] = 1
        return _aux_var
    else:
        _aux_var = _get_aux_var(parent_block, pe.sin(arg))
        arg = replace_sub_expression_with_aux_var(arg, parent_block)
        relaxation_side = relaxation_side_map[node]
        degree_map[_aux_var] = 1
        relaxation = PWSinRelaxation()
        relaxation.set_input(x=arg, aux_var=_aux_var, relaxation_side=relaxation_side)
        aux_var_map[id(arg), 'sin'] = (_aux_var, relaxation)
        setattr(parent_block.relaxations, 'rel'+str(counter), relaxation)
        counter.increment()
        return _aux_var


def _relax_leaf_to_root_cos(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg = values[0]
    degree = degree_map[arg]
    if degree == 0:
        res = pe.cos(arg)
        degree_map[res] = 0
        return res
    elif (id(arg), 'cos') in aux_var_map:
        _aux_var, relaxation = aux_var_map[id(arg), 'cos']
        relaxation_side = relaxation_side_map[node]
        if relaxation_side != relaxation.relaxation_side:
            relaxation.relaxation_side = RelaxationSide.BOTH
        degree_map[_aux_var] = 1
        return _aux_var
    else:
        _aux_var = _get_aux_var(parent_block, pe.cos(arg))
        arg = replace_sub_expression_with_aux_var(arg, parent_block)
        relaxation_side = relaxation_side_map[node]
        degree_map[_aux_var] = 1
        relaxation = PWCosRelaxation()
        relaxation.set_input(x=arg, aux_var=_aux_var, relaxation_side=relaxation_side)
        aux_var_map[id(arg), 'cos'] = (_aux_var, relaxation)
        setattr(parent_block.relaxations, 'rel'+str(counter), relaxation)
        counter.increment()
        return _aux_var


def _relax_leaf_to_root_arctan(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg = values[0]
    degree = degree_map[arg]
    if degree == 0:
        res = pe.atan(arg)
        degree_map[res] = 0
        return res
    elif (id(arg), 'arctan') in aux_var_map:
        _aux_var, relaxation = aux_var_map[id(arg), 'arctan']
        relaxation_side = relaxation_side_map[node]
        if relaxation_side != relaxation.relaxation_side:
            relaxation.relaxation_side = RelaxationSide.BOTH
        degree_map[_aux_var] = 1
        return _aux_var
    else:
        _aux_var = _get_aux_var(parent_block, pe.atan(arg))
        arg = replace_sub_expression_with_aux_var(arg, parent_block)
        relaxation_side = relaxation_side_map[node]
        degree_map[_aux_var] = 1
        relaxation = PWArctanRelaxation()
        relaxation.set_input(x=arg, aux_var=_aux_var, relaxation_side=relaxation_side)
        aux_var_map[id(arg), 'arctan'] = (_aux_var, relaxation)
        setattr(parent_block.relaxations, 'rel'+str(counter), relaxation)
        counter.increment()
        return _aux_var


_unary_leaf_to_root_map = dict()
_unary_leaf_to_root_map['exp'] = _relax_leaf_to_root_exp
_unary_leaf_to_root_map['log'] = _relax_leaf_to_root_log
_unary_leaf_to_root_map['log10'] = _relax_leaf_to_root_log10
_unary_leaf_to_root_map['sin'] = _relax_leaf_to_root_sin
_unary_leaf_to_root_map['cos'] = _relax_leaf_to_root_cos
_unary_leaf_to_root_map['atan'] = _relax_leaf_to_root_arctan
_unary_leaf_to_root_map['sqrt'] = _relax_leaf_to_root_sqrt


def _relax_leaf_to_root_UnaryFunctionExpression(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    if node.getname() in _unary_leaf_to_root_map:
        return _unary_leaf_to_root_map[node.getname()](node=node, values=values, aux_var_map=aux_var_map,
                                                       degree_map=degree_map, parent_block=parent_block,
                                                       relaxation_side_map=relaxation_side_map, counter=counter)
    else:
        raise NotImplementedError('Cannot automatically relax ' + str(node))


def _relax_leaf_to_root_GeneralExpression(node, values, aux_var_map, degree_map, parent_block, relaxation_side_map, counter):
    arg = values[0]
    return arg


_relax_leaf_to_root_map = dict()
_relax_leaf_to_root_map[numeric_expr.ProductExpression] = _relax_leaf_to_root_ProductExpression
_relax_leaf_to_root_map[numeric_expr.SumExpression] = _relax_leaf_to_root_SumExpression
_relax_leaf_to_root_map[numeric_expr.MonomialTermExpression] = _relax_leaf_to_root_ProductExpression
_relax_leaf_to_root_map[numeric_expr.NegationExpression] = _relax_leaf_to_root_NegationExpression
_relax_leaf_to_root_map[numeric_expr.PowExpression] = _relax_leaf_to_root_PowExpression
_relax_leaf_to_root_map[numeric_expr.DivisionExpression] = _relax_leaf_to_root_DivisionExpression
_relax_leaf_to_root_map[numeric_expr.UnaryFunctionExpression] = _relax_leaf_to_root_UnaryFunctionExpression
_relax_leaf_to_root_map[numeric_expr.NPV_ProductExpression] = _relax_leaf_to_root_ProductExpression
_relax_leaf_to_root_map[numeric_expr.NPV_SumExpression] = _relax_leaf_to_root_SumExpression
_relax_leaf_to_root_map[numeric_expr.NPV_NegationExpression] = _relax_leaf_to_root_NegationExpression
_relax_leaf_to_root_map[numeric_expr.NPV_PowExpression] = _relax_leaf_to_root_PowExpression
_relax_leaf_to_root_map[numeric_expr.NPV_DivisionExpression] = _relax_leaf_to_root_DivisionExpression
_relax_leaf_to_root_map[numeric_expr.NPV_UnaryFunctionExpression] = _relax_leaf_to_root_UnaryFunctionExpression
_relax_leaf_to_root_map[_GeneralExpressionData] = _relax_leaf_to_root_GeneralExpression
_relax_leaf_to_root_map[SimpleExpression] = _relax_leaf_to_root_GeneralExpression


def _relax_root_to_leaf_ProductExpression(node, relaxation_side_map):
    arg1, arg2 = node.args
    if is_fixed(arg1):
        relaxation_side_map[arg1] = RelaxationSide.BOTH
        if isinstance(arg1, numeric_expr.ProductExpression):  # see Pyomo issue #1147
            arg1_arg1 = arg1.args[0]
            arg1_arg2 = arg1.args[1]
            try:
                arg1_arg1_val = pe.value(arg1_arg1)
            except ValueError:
                arg1_arg1_val = None
            try:
                arg1_arg2_val = pe.value(arg1_arg2)
            except ValueError:
                arg1_arg2_val = None
            if arg1_arg1_val == 0 or arg1_arg2_val == 0:
                arg1_val = 0
            else:
                arg1_val = pe.value(arg1)
        else:
            arg1_val = pe.value(arg1)
        if arg1_val >= 0:
            relaxation_side_map[arg2] = relaxation_side_map[node]
        else:
            if relaxation_side_map[node] == RelaxationSide.UNDER:
                relaxation_side_map[arg2] = RelaxationSide.OVER
            elif relaxation_side_map[node] == RelaxationSide.OVER:
                relaxation_side_map[arg2] = RelaxationSide.UNDER
            else:
                relaxation_side_map[arg2] = RelaxationSide.BOTH
    elif is_fixed(arg2):
        relaxation_side_map[arg2] = RelaxationSide.BOTH
        if isinstance(arg2, numeric_expr.ProductExpression):  # see Pyomo issue #1147
            arg2_arg1 = arg2.args[0]
            arg2_arg2 = arg2.args[1]
            try:
                arg2_arg1_val = pe.value(arg2_arg1)
            except ValueError:
                arg2_arg1_val = None
            try:
                arg2_arg2_val = pe.value(arg2_arg2)
            except ValueError:
                arg2_arg2_val = None
            if arg2_arg1_val == 0 or arg2_arg2_val == 0:
                arg2_val = 0
            else:
                arg2_val = pe.value(arg2)
        else:
            arg2_val = pe.value(arg2)
        if arg2_val >= 0:
            relaxation_side_map[arg1] = relaxation_side_map[node]
        else:
            if relaxation_side_map[node] == RelaxationSide.UNDER:
                relaxation_side_map[arg1] = RelaxationSide.OVER
            elif relaxation_side_map[node] == RelaxationSide.OVER:
                relaxation_side_map[arg1] = RelaxationSide.UNDER
            else:
                relaxation_side_map[arg1] = RelaxationSide.BOTH
    else:
        relaxation_side_map[arg1] = RelaxationSide.BOTH
        relaxation_side_map[arg2] = RelaxationSide.BOTH


def _relax_root_to_leaf_DivisionExpression(node, relaxation_side_map):
    arg1, arg2 = node.args
    relaxation_side_map[arg1] = RelaxationSide.BOTH
    relaxation_side_map[arg2] = RelaxationSide.BOTH


def _relax_root_to_leaf_SumExpression(node, relaxation_side_map):
    relaxation_side = relaxation_side_map[node]

    for arg in node.args:
        relaxation_side_map[arg] = relaxation_side


def _relax_root_to_leaf_NegationExpression(node, relaxation_side_map):
    arg = node.args[0]
    relaxation_side = relaxation_side_map[node]
    if relaxation_side == RelaxationSide.BOTH:
        relaxation_side_map[arg] = RelaxationSide.BOTH
    elif relaxation_side == RelaxationSide.UNDER:
        relaxation_side_map[arg] = RelaxationSide.OVER
    else:
        assert relaxation_side == RelaxationSide.OVER
        relaxation_side_map[arg] = RelaxationSide.UNDER


def _relax_root_to_leaf_PowExpression(node, relaxation_side_map):
    arg1, arg2 = node.args
    relaxation_side_map[arg1] = RelaxationSide.BOTH
    relaxation_side_map[arg2] = RelaxationSide.BOTH


def _relax_root_to_leaf_sqrt(node, relaxation_side_map):
    arg = node.args[0]
    relaxation_side_map[arg] = relaxation_side_map[node]


def _relax_root_to_leaf_exp(node, relaxation_side_map):
    arg = node.args[0]
    relaxation_side_map[arg] = relaxation_side_map[node]


def _relax_root_to_leaf_log(node, relaxation_side_map):
    arg = node.args[0]
    relaxation_side_map[arg] = relaxation_side_map[node]


def _relax_root_to_leaf_log10(node, relaxation_side_map):
    arg = node.args[0]
    relaxation_side_map[arg] = relaxation_side_map[node]


def _relax_root_to_leaf_sin(node, relaxation_side_map):
    arg = node.args[0]
    relaxation_side_map[arg] = RelaxationSide.BOTH


def _relax_root_to_leaf_cos(node, relaxation_side_map):
    arg = node.args[0]
    relaxation_side_map[arg] = RelaxationSide.BOTH


def _relax_root_to_leaf_arctan(node, relaxation_side_map):
    arg = node.args[0]
    relaxation_side_map[arg] = RelaxationSide.BOTH


_unary_root_to_leaf_map = dict()
_unary_root_to_leaf_map['exp'] = _relax_root_to_leaf_exp
_unary_root_to_leaf_map['log'] = _relax_root_to_leaf_log
_unary_root_to_leaf_map['log10'] = _relax_root_to_leaf_log10
_unary_root_to_leaf_map['sin'] = _relax_root_to_leaf_sin
_unary_root_to_leaf_map['cos'] = _relax_root_to_leaf_cos
_unary_root_to_leaf_map['atan'] = _relax_root_to_leaf_arctan
_unary_root_to_leaf_map['sqrt'] = _relax_root_to_leaf_sqrt


def _relax_root_to_leaf_UnaryFunctionExpression(node, relaxation_side_map):
    if node.getname() in _unary_root_to_leaf_map:
        _unary_root_to_leaf_map[node.getname()](node, relaxation_side_map)
    else:
        raise NotImplementedError('Cannot automatically relax ' + str(node))


def _relax_root_to_leaf_GeneralExpression(node, relaxation_side_map):
    relaxation_side = relaxation_side_map[node]
    relaxation_side_map[node.expr] = relaxation_side


_relax_root_to_leaf_map = dict()
_relax_root_to_leaf_map[numeric_expr.ProductExpression] = _relax_root_to_leaf_ProductExpression
_relax_root_to_leaf_map[numeric_expr.SumExpression] = _relax_root_to_leaf_SumExpression
_relax_root_to_leaf_map[numeric_expr.MonomialTermExpression] = _relax_root_to_leaf_ProductExpression
_relax_root_to_leaf_map[numeric_expr.NegationExpression] = _relax_root_to_leaf_NegationExpression
_relax_root_to_leaf_map[numeric_expr.PowExpression] = _relax_root_to_leaf_PowExpression
_relax_root_to_leaf_map[numeric_expr.DivisionExpression] = _relax_root_to_leaf_DivisionExpression
_relax_root_to_leaf_map[numeric_expr.UnaryFunctionExpression] = _relax_root_to_leaf_UnaryFunctionExpression
_relax_root_to_leaf_map[numeric_expr.NPV_ProductExpression] = _relax_root_to_leaf_ProductExpression
_relax_root_to_leaf_map[numeric_expr.NPV_SumExpression] = _relax_root_to_leaf_SumExpression
_relax_root_to_leaf_map[numeric_expr.NPV_NegationExpression] = _relax_root_to_leaf_NegationExpression
_relax_root_to_leaf_map[numeric_expr.NPV_PowExpression] = _relax_root_to_leaf_PowExpression
_relax_root_to_leaf_map[numeric_expr.NPV_DivisionExpression] = _relax_root_to_leaf_DivisionExpression
_relax_root_to_leaf_map[numeric_expr.NPV_UnaryFunctionExpression] = _relax_root_to_leaf_UnaryFunctionExpression
_relax_root_to_leaf_map[_GeneralExpressionData] = _relax_root_to_leaf_GeneralExpression
_relax_root_to_leaf_map[SimpleExpression] = _relax_root_to_leaf_GeneralExpression


class _FactorableRelaxationVisitor(ExpressionValueVisitor):
    """
    This walker generates new constraints with nonlinear terms replaced by
    auxiliary variables, and relaxations relating the auxilliary variables to
    the original variables.
    """
    def __init__(self, aux_var_map, parent_block, relaxation_side_map, counter, degree_map):
        self.aux_var_map = aux_var_map
        self.parent_block = parent_block
        self.relaxation_side_map = relaxation_side_map
        self.counter = counter
        self.degree_map = degree_map

    def visit(self, node, values):
        if node.__class__ in _relax_leaf_to_root_map:
            res = _relax_leaf_to_root_map[node.__class__](node, values, self.aux_var_map, self.degree_map,
                                                          self.parent_block, self.relaxation_side_map, self.counter)
            return res
        else:
            raise NotImplementedError('Cannot relax an expression of type ' + str(type(node)))

    def visiting_potential_leaf(self, node):
        if node.__class__ in nonpyomo_leaf_types:
            self.degree_map[node] = 0
            return True, node

        if node.is_variable_type():
            if node.fixed:
                self.degree_map[node] = 0
            else:
                self.degree_map[node] = 1
            return True, node

        if not node.is_expression_type():
            self.degree_map[node] = 0
            return True, node

        if node.__class__ in _relax_root_to_leaf_map:
            _relax_root_to_leaf_map[node.__class__](node, self.relaxation_side_map)
        else:
            raise NotImplementedError('Cannot relax an expression of type ' + str(type(node)))

        return False, None


def _get_prefix_notation(expr):
    pn = convert_expression_to_prefix_notation(expr, include_named_exprs=False)
    res = list()
    for i in pn:
        itype = type(i)
        if itype is tuple or itype in nonpyomo_leaf_types:
            res.append(i)
        elif isinstance(i, NumericValue):
            if i.is_fixed():
                res.append(pe.value(i))
            else:
                assert i.is_variable_type()
                res.append(id(i))
        else:
            raise NotImplementedError(f'unexpected entry in prefix notation: {str(i)}')
    return tuple(res)


def _relax_expr(expr, aux_var_map, parent_block, relaxation_side_map, counter, degree_map):
    visitor = _FactorableRelaxationVisitor(aux_var_map=aux_var_map, parent_block=parent_block,
                                           relaxation_side_map=relaxation_side_map, counter=counter,
                                           degree_map=degree_map)
    new_expr = visitor.dfs_postorder_stack(expr)
    return new_expr


def _relax_split_expr(
    expr: ExpressionBase,
    aux_var_map: MutableMapping[
        Tuple,
        Tuple[NumericValue,
              Union[BaseRelaxationData,
                    Tuple[BaseRelaxationData, BaseRelaxationData]]]
    ],
    parent_block: _BlockData,
    relaxation_side_map: MutableMapping[NumericValue, RelaxationSide],
    counter: RelaxationCounter,
    degree_map: MutableMapping[NumericValue, int],
    eigenvalue_bounder: EigenValueBounder,
    max_vars_per_alpha_bb: int,
    max_eigenvalue_for_alpha_bb: float,
    eigenvalue_opt: Optional[appsi.base.Solver],
) -> NumericValue:
    relaxation_side = relaxation_side_map[expr]
    hessian = Hessian(expr, opt=eigenvalue_opt, method=eigenvalue_bounder)
    vlist = hessian.variables()
    min_eig = hessian.get_minimum_eigenvalue()
    max_eig = hessian.get_maximum_eigenvalue()
    is_convex = min_eig >= 0
    is_concave = max_eig <= 0

    all_vars_bounded = True
    for v in vlist:
        v_lb, v_ub = v.bounds
        if v_lb is None or v_ub is None:
            all_vars_bounded = False
            break

    if len(vlist) == 1 and (is_convex or is_concave):
        pn = _get_prefix_notation(expr)
        if pn in aux_var_map:
            new_expr, relaxation = aux_var_map[pn]
            if relaxation_side != relaxation.relaxation_side:
                relaxation.relaxation_side = RelaxationSide.BOTH
        else:
            new_expr = _get_aux_var(parent_block, expr)
            relaxation = PWUnivariateRelaxation()
            if is_convex:
                shape = FunctionShape.CONVEX
            else:
                shape = FunctionShape.CONCAVE
            relaxation.set_input(
                x=vlist[0], aux_var=new_expr, relaxation_side=relaxation_side,
                f_x_expr=expr, shape=shape,
            )
            aux_var_map[pn] = (new_expr, relaxation)
            setattr(parent_block.relaxations, 'rel' + str(counter), relaxation)
            counter.increment()
            degree_map[new_expr] = 1
    elif ((is_convex and relaxation_side == RelaxationSide.UNDER)
          or (is_concave and relaxation_side == RelaxationSide.OVER)):
        pn = _get_prefix_notation(expr)
        if pn in aux_var_map:
            new_expr, (underestimator, overestimator) = aux_var_map[pn]
        else:
            new_expr, underestimator, overestimator = None, None, None
        if new_expr is None:
            new_expr = _get_aux_var(parent_block, expr)
        if (
            (is_convex and underestimator is None)
            or (is_concave and overestimator is None)
        ):
            relaxation = MultivariateRelaxation()
            if is_convex:
                shape = FunctionShape.CONVEX
                underestimator = relaxation
            else:
                shape = FunctionShape.CONCAVE
                overestimator = relaxation
            relaxation.set_input(
                aux_var=new_expr, shape=shape, f_x_expr=expr,
            )
            aux_var_map[pn] = (new_expr, (underestimator, overestimator))
            setattr(parent_block.relaxations, 'rel' + str(counter), relaxation)
            counter.increment()
            degree_map[new_expr] = 1
    elif (
        all_vars_bounded
        and len(vlist) <= max_vars_per_alpha_bb
        and (
            (relaxation_side == RelaxationSide.UNDER and min_eig >= -abs(max_eigenvalue_for_alpha_bb))
            or (relaxation_side == RelaxationSide.OVER and max_eig <= abs(max_eigenvalue_for_alpha_bb))
        )
    ):
        pn = _get_prefix_notation(expr)
        if pn in aux_var_map:
            new_expr, (underestimator, overestimator) = aux_var_map[pn]
        else:
            new_expr, underestimator, overestimator = None, None, None
        if new_expr is None:
            new_expr = _get_aux_var(parent_block, expr)
        if (
            (relaxation_side == RelaxationSide.UNDER and underestimator is None)
            or (relaxation_side == RelaxationSide.OVER and overestimator is None)
        ):
            relaxation = AlphaBBRelaxation()
            relaxation.set_input(
                aux_var=new_expr,
                f_x_expr=expr,
                relaxation_side=relaxation_side,
                hessian=hessian,
            )
            if relaxation_side == RelaxationSide.UNDER:
                underestimator = relaxation
            else:
                overestimator = relaxation
            aux_var_map[pn] = (new_expr, (underestimator, overestimator))
            setattr(parent_block.relaxations, 'rel' + str(counter), relaxation)
            counter.increment()
            degree_map[new_expr] = 1
    else:
        visitor = _FactorableRelaxationVisitor(aux_var_map=aux_var_map,
                                               parent_block=parent_block,
                                               relaxation_side_map=relaxation_side_map,
                                               counter=counter,
                                               degree_map=degree_map)
        new_expr = visitor.dfs_postorder_stack(expr)
    return new_expr


def _relax_expr_with_convexity_check(
    orig_expr: ExpressionBase,
    aux_var_map: MutableMapping[
        Tuple,
        Tuple[NumericValue,
              Union[BaseRelaxationData,
                    Tuple[BaseRelaxationData, BaseRelaxationData]]]
    ],
    parent_block: _BlockData,
    relaxation_side_map: MutableMapping[NumericValue, RelaxationSide],
    counter: RelaxationCounter,
    degree_map: MutableMapping[NumericValue, int],
    perform_expression_simplification: bool,
    eigenvalue_bounder: EigenValueBounder,
    max_vars_per_alpha_bb: int,
    max_eigenvalue_for_alpha_bb: float,
    eigenvalue_opt: Optional[appsi.base.Solver],
):
    if relaxation_side_map[orig_expr] == RelaxationSide.BOTH:
        res_list = []
        for side in [RelaxationSide.UNDER, RelaxationSide.OVER]:
            relaxation_side_map[orig_expr] = side
            tmp_res = _relax_expr_with_convexity_check(
                orig_expr=orig_expr,
                aux_var_map=aux_var_map,
                parent_block=parent_block,
                relaxation_side_map=relaxation_side_map,
                counter=counter,
                degree_map=degree_map,
                perform_expression_simplification=perform_expression_simplification,
                eigenvalue_bounder=eigenvalue_bounder,
                max_vars_per_alpha_bb=max_vars_per_alpha_bb,
                max_eigenvalue_for_alpha_bb=max_eigenvalue_for_alpha_bb,
                eigenvalue_opt=eigenvalue_opt,
            )
            res_list.append(tmp_res)
        linking_expr = res_list[0] - res_list[1]
        linking_repn = generate_standard_repn(linking_expr, compute_values=False, quadratic=True)
        linking_expr = linking_repn.to_expression()
        if is_constant(linking_expr):
            assert value(linking_expr) == 0
        else:
            parent_block.aux_cons.add(linking_repn.to_expression() == 0)
        res = res_list[0]
        relaxation_side_map[orig_expr] = RelaxationSide.BOTH
    else:
        if perform_expression_simplification:
            _expr = simplify_expr(orig_expr)
        else:
            _expr = orig_expr
        list_of_exprs = split_expr(_expr)
        list_of_new_exprs = list()

        for expr in list_of_exprs:
            relaxation_side_map[expr] = relaxation_side_map[orig_expr]
            new_expr = _relax_split_expr(
                expr=expr,
                aux_var_map=aux_var_map,
                parent_block=parent_block,
                relaxation_side_map=relaxation_side_map,
                counter=counter,
                degree_map=degree_map,
                eigenvalue_bounder=eigenvalue_bounder,
                max_vars_per_alpha_bb=max_vars_per_alpha_bb,
                max_eigenvalue_for_alpha_bb=max_eigenvalue_for_alpha_bb,
                eigenvalue_opt=eigenvalue_opt,
            )
            list_of_new_exprs.append(new_expr)
        res = sum(list_of_new_exprs)
    return res


def relax(
    model,
    descend_into=None,
    in_place=False,
    use_fbbt=True,
    fbbt_options=None,
    perform_expression_simplification: bool = True,
    use_alpha_bb: bool = True,
    eigenvalue_bounder: EigenValueBounder = EigenValueBounder.GershgorinWithSimplification,
    max_vars_per_alpha_bb: int = 4,
    max_eigenvalue_for_alpha_bb: float = 100,
    eigenvalue_opt: Optional[appsi.base.Solver] = None,
):
    """
    Create a convex relaxation of the model.

    Parameters
    ----------
    model: pyomo.core.base.block._BlockData or pyomo.core.base.PyomoModel.ConcreteModel
        The model or block to be relaxed
    descend_into: type or tuple of type, optional
        The types of pyomo components that should be checked for constraints to be relaxed. The
        default is (Block, Disjunct).
    in_place: bool, optional
        If False (default=False), model will be cloned, and the clone will be relaxed. 
        If True, then model will be modified in place.
    use_fbbt: bool, optional
        If True (default=True), then FBBT will be used to tighten variable bounds. If False, 
        FBBT will not be used.
    fbbt_options: dict, optional
        The options to pass to the call to fbbt. See pyomo.contrib.fbbt.fbbt.fbbt for details.
    convexity_effort: ConvexityEffort

    Returns
    -------
    m: pyomo.core.base.block._BlockData or pyomo.core.base.PyomoModel.ConcreteModel
        The relaxed model
    """
    """
    For now, we will use FBBT both before relaxing the model and after relaxing the model. The reason we need to 
    do it before relaxing the model is that the variable bounds will affect the structure of the relaxation. For 
    example, if we need to relax x**3 and x >= 0, then we know x**3 is convex, and we can relax it as a 
    convex, univariate function. However, if x can be positive or negative, then x**3 is neither convex nor concave.
    In this case, we relax it by reformulating it as x * x**2. The hope is that performing FBBT before relaxing 
    the model will help identify things like x >= 0 and therefore x**3 is convex. The correct way to do this is to 
    update the relaxation classes so that the original expression is known, and the best relaxation can be used 
    anytime the variable bounds are updated. For example, suppose the model is relaxed and, only after OBBT is 
    performed, we find out x >= 0. We should be able to easily update the relaxation so that x**3 is then relaxed 
    as a convex univariate function. The reason FBBT needs to be performed after relaxing the model is that 
    we want to make sure that all of the auxilliary variables introduced get tightened bounds. The correct way to 
    handle this is to perform FBBT with the original model with suspect, which forms a DAG. Each auxilliary variable 
    introduced in the relaxed model corresponds to a node in the DAG. If we use suspect, then we can easily 
    update the bounds of the auxilliary variables without performing FBBT a second time.
    """
    if not in_place:
        m = model.clone()
    else:
        m = model

    if fbbt_options is None:
        fbbt_options = dict()

    if use_fbbt:
        it = appsi.fbbt.IntervalTightener()
        for k, v in fbbt_options.items():
            setattr(it.config, k, v)
        it.perform_fbbt(m)

    if descend_into is None:
        descend_into = (pe.Block, Disjunct)

    aux_var_map = dict()
    counter_dict = dict()
    degree_map = ComponentMap()

    for c in nonrelaxation_component_data_objects(m, ctype=Constraint, active=True, descend_into=descend_into, sort=True):
        body_degree = polynomial_degree(c.body)
        if body_degree is not None:
            if body_degree <= 1:
                continue

        if c.lower is not None and c.upper is not None:
            relaxation_side = RelaxationSide.BOTH
        elif c.lower is not None:
            relaxation_side = RelaxationSide.OVER
        elif c.upper is not None:
            relaxation_side = RelaxationSide.UNDER
        else:
            raise ValueError('Encountered a constraint without a lower or an upper bound: ' + str(c))

        parent_block = c.parent_block()

        if parent_block in counter_dict:
            counter = counter_dict[parent_block]
        else:
            parent_block.relaxations = pe.Block()
            parent_block.aux_vars = pe.VarList()
            parent_block.aux_cons = pe.ConstraintList()
            counter = RelaxationCounter()
            counter_dict[parent_block] = counter

        repn = generate_standard_repn(c.body, quadratic=False, compute_values=False)
        assert len(repn.quadratic_vars) == 0
        assert repn.nonlinear_expr is not None
        if len(repn.linear_vars) > 0:
            new_body = numeric_expr.LinearExpression(constant=repn.constant, linear_coefs=repn.linear_coefs, linear_vars=repn.linear_vars)
        else:
            new_body = repn.constant

        relaxation_side_map = ComponentMap()
        relaxation_side_map[repn.nonlinear_expr] = relaxation_side

        if not use_alpha_bb:
            new_body += _relax_expr(
                expr=repn.nonlinear_expr, aux_var_map=aux_var_map,
                parent_block=parent_block, relaxation_side_map=relaxation_side_map,
                counter=counter, degree_map=degree_map
            )
        else:
            new_body += _relax_expr_with_convexity_check(
                orig_expr=repn.nonlinear_expr, aux_var_map=aux_var_map,
                parent_block=parent_block, relaxation_side_map=relaxation_side_map,
                counter=counter, degree_map=degree_map,
                perform_expression_simplification=perform_expression_simplification,
                eigenvalue_bounder=eigenvalue_bounder,
                max_vars_per_alpha_bb=max_vars_per_alpha_bb,
                max_eigenvalue_for_alpha_bb=max_eigenvalue_for_alpha_bb,
                eigenvalue_opt=eigenvalue_opt,
            )
        lb = c.lower
        ub = c.upper
        parent_block.aux_cons.add(pe.inequality(lb, new_body, ub))
        parent_component = c.parent_component()
        if parent_component.is_indexed():
            del parent_component[c.index()]
        else:
            parent_block.del_component(c)

    for c in nonrelaxation_component_data_objects(m, ctype=pe.Objective, active=True, descend_into=descend_into, sort=True):
        degree = polynomial_degree(c.expr)
        if degree is not None:
            if degree <= 1:
                continue

        if c.sense == pe.minimize:
            relaxation_side = RelaxationSide.UNDER
        elif c.sense == pe.maximize:
            relaxation_side = RelaxationSide.OVER
        else:
            raise ValueError('Encountered an objective with an unrecognized sense: ' + str(c))

        parent_block = c.parent_block()

        if parent_block in counter_dict:
            counter = counter_dict[parent_block]
        else:
            parent_block.relaxations = pe.Block()
            parent_block.aux_vars = pe.VarList()
            parent_block.aux_cons = pe.ConstraintList()
            counter = RelaxationCounter()
            counter_dict[parent_block] = counter

        if not hasattr(parent_block, 'aux_objectives'):
            parent_block.aux_objectives = pe.ObjectiveList()

        repn = generate_standard_repn(c.expr, quadratic=False, compute_values=False)
        assert len(repn.quadratic_vars) == 0
        assert repn.nonlinear_expr is not None
        if len(repn.linear_vars) > 0:
            new_body = numeric_expr.LinearExpression(constant=repn.constant, linear_coefs=repn.linear_coefs, linear_vars=repn.linear_vars)
        else:
            new_body = repn.constant

        relaxation_side_map = ComponentMap()
        relaxation_side_map[repn.nonlinear_expr] = relaxation_side

        if not use_alpha_bb:
            new_body += _relax_expr(
                expr=repn.nonlinear_expr, aux_var_map=aux_var_map,
                parent_block=parent_block, relaxation_side_map=relaxation_side_map,
                counter=counter, degree_map=degree_map
            )
        else:
            new_body += _relax_expr_with_convexity_check(
                orig_expr=repn.nonlinear_expr, aux_var_map=aux_var_map,
                parent_block=parent_block, relaxation_side_map=relaxation_side_map,
                counter=counter, degree_map=degree_map,
                perform_expression_simplification=perform_expression_simplification,
                eigenvalue_bounder=eigenvalue_bounder,
                max_vars_per_alpha_bb=max_vars_per_alpha_bb,
                max_eigenvalue_for_alpha_bb=max_eigenvalue_for_alpha_bb,
                eigenvalue_opt=eigenvalue_opt,
            )
        sense = c.sense
        parent_block.aux_objectives.add(new_body, sense=sense)
        parent_component = c.parent_component()
        if parent_component.is_indexed():
            del parent_component[c.index()]
        else:
            parent_block.del_component(c)

    if use_fbbt:
        for relaxation in relaxation_data_objects(m, descend_into=True, active=True):
            relaxation.rebuild(build_nonlinear_constraint=True)

        it = appsi.fbbt.IntervalTightener()
        for k, v in fbbt_options.items():
            setattr(it.config, k, v)
        it.config.deactivate_satisfied_constraints = False
        it.perform_fbbt(m)

        for relaxation in relaxation_data_objects(m, descend_into=True, active=True):
            relaxation.use_linear_relaxation = True
            relaxation.rebuild()
    else:
        for relaxation in relaxation_data_objects(m, descend_into=True, active=True):
            relaxation.use_linear_relaxation = True
            relaxation.rebuild()

    return m



import sys
if sys.version_info.major != 3 or sys.version_info.minor < 6:
    raise EnvironmentError('Coramin only supports Python 3.6 and newer.')

from coramin import utils
from coramin import domain_reduction
from coramin import relaxations
from coramin import algorithms
from coramin import third_party
from .utils import (
    RelaxationSide,
    FunctionShape,
    Effort,
    EigenValueBounder,
    simplify_expr,
    get_objective
)



import numpy as np
try:
    import plotly.graph_objects as go
except ImportError:
    pass
import pyomo.environ as pe
from .pyomo_utils import get_objective
from .coramin_enums import RelaxationSide
from pyomo.solvers.plugins.solvers.persistent_solver import PersistentSolver
try:
    import tqdm
except ImportError:
    tqdm = None


def _solve(m, using_persistent_solver, solver, rhs_vars, aux_var, obj):
    obj.activate()
    if using_persistent_solver:
        for v in rhs_vars:
            solver.update_var(v)
        solver.set_objective(obj)
        res = solver.solve(load_solutions=False, save_results=False)
    else:
        res = solver.solve(m, load_solutions=False)
    if res.solver.termination_condition != pe.TerminationCondition.optimal:
        raise RuntimeError('Could not produce plot because solver did not terminate optimally')
    if using_persistent_solver:
        solver.load_vars([aux_var])
    else:
        m.solutions.load_from(res)
    obj.deactivate()


def _solve_loop(m, x, w, x_list, using_persistent_solver, solver):
    w_list = list()
    for _xval in x_list:
        x.fix(_xval)
        if using_persistent_solver:
            solver.update_var(x)
            res = solver.solve(load_solutions=False, save_results=False)
        else:
            res = solver.solve(m, load_solutions=False)
        if res.solver.termination_condition != pe.TerminationCondition.optimal:
            raise RuntimeError(
                'Could not produce plot because solver did not terminate optimally. Termination condition: ' + str(
                    res.solver.termination_condition))
        if using_persistent_solver:
            solver.load_vars([w])
        else:
            m.solutions.load_from(res)
        w_list.append(w.value)
    return w_list


def _plot_2d(m, relaxation, solver, show_plot, num_pts):
    using_persistent_solver = isinstance(solver, PersistentSolver)

    x = relaxation.get_rhs_vars()[0]
    w = relaxation.get_aux_var()

    if not x.has_lb() or not x.has_ub():
        raise ValueError('rhs var must have bounds')

    orig_xval = x.value
    orig_wval = w.value
    xlb = pe.value(x.lb)
    xub = pe.value(x.ub)

    orig_obj = get_objective(m)
    if orig_obj is not None:
        orig_obj.deactivate()

    x_list = np.linspace(xlb, xub, num_pts)
    x_list = [float(i) for i in x_list]
    w_true = list()

    rhs_expr = relaxation.get_rhs_expr()
    for _x in x_list:
        x.value = float(_x)
        w_true.append(pe.value(rhs_expr))
    plotly_data = [go.Scatter(x=x_list, y=w_true, name=str(rhs_expr))]

    m._plotting_objective = pe.Objective(expr=w)
    if using_persistent_solver:
        solver.set_instance(m)

    if relaxation.relaxation_side in {RelaxationSide.UNDER, RelaxationSide.BOTH}:
        w_min = _solve_loop(m, x, w, x_list, using_persistent_solver, solver)
        plotly_data.append(go.Scatter(x=x_list, y=w_min, name='underestimator'))

    del m._plotting_objective
    m._plotting_objective = pe.Objective(expr=w, sense=pe.maximize)
    if using_persistent_solver:
        solver.set_objective(m._plotting_objective)

    if relaxation.relaxation_side in {RelaxationSide.OVER, RelaxationSide.BOTH}:
        w_max = _solve_loop(m, x, w, x_list, using_persistent_solver, solver)
        plotly_data.append(go.Scatter(x=x_list, y=w_max, name='overestimator'))

    fig = go.Figure(data=plotly_data)
    if show_plot:
        fig.show()

    x.unfix()
    x.value = orig_xval
    w.value = orig_wval
    del m._plotting_objective
    if orig_obj is not None:
        orig_obj.activate()


def _plot_3d(m, relaxation, solver, show_plot, num_pts):
    using_persistent_solver = isinstance(solver, PersistentSolver)

    rhs_vars = relaxation.get_rhs_vars()
    x, y = rhs_vars
    w = relaxation.get_aux_var()

    if not x.has_lb() or not x.has_ub() or not y.has_lb() or not y.has_ub():
        raise ValueError('rhs vars must have bounds')

    orig_xval = x.value
    orig_yval = y.value
    orig_wval = w.value

    orig_obj = get_objective(m)
    if orig_obj is not None:
        orig_obj.deactivate()

    m._underestimator_obj = pe.Objective(expr=w)
    m._overestimator_obj = pe.Objective(expr=w, sense=pe.maximize)
    m._underestimator_obj.deactivate()
    m._overestimator_obj.deactivate()
    if using_persistent_solver:
        solver.set_instance(m)

    x_list = np.linspace(x.lb, x.ub, num_pts)
    y_list = np.linspace(y.lb, y.ub, num_pts)
    x_list = [float(i) for i in x_list]
    y_list = [float(i) for i in y_list]
    w_true = np.empty((num_pts, num_pts), dtype=np.double)
    w_min = np.empty((num_pts, num_pts), dtype=np.double)
    w_max = np.empty((num_pts, num_pts), dtype=np.double)

    rhs_expr = relaxation.get_rhs_expr()

    def sub_loop(x_ndx, _x):
        x.fix(_x)
        for y_ndx, _y in enumerate(y_list):
            y.fix(_y)
            w_true[x_ndx, y_ndx] = pe.value(rhs_expr)
            if relaxation.relaxation_side in {RelaxationSide.UNDER, RelaxationSide.BOTH}:
                _solve(m, using_persistent_solver, solver, rhs_vars, w, m._underestimator_obj)
                w_min[x_ndx, y_ndx] = w.value
            if relaxation.relaxation_side in {RelaxationSide.OVER, RelaxationSide.BOTH}:
                _solve(m, using_persistent_solver, solver, rhs_vars, w, m._overestimator_obj)
                w_max[x_ndx, y_ndx] = w.value

    if tqdm is not None:
        for x_ndx, _x in tqdm.tqdm(list(enumerate(x_list))):
            sub_loop(x_ndx, _x)
    else:
        for x_ndx, _x in enumerate(x_list):
            sub_loop(x_ndx, _x)

    plotly_data = list()
    plotly_data.append(go.Surface(x=x_list, y=y_list, z=w_true, name=str(rhs_expr)))
    if relaxation.relaxation_side in {RelaxationSide.UNDER, RelaxationSide.BOTH}:
        plotly_data.append(go.Surface(x=x_list, y=y_list, z=w_min, name='underestimator'))
    if relaxation.relaxation_side in {RelaxationSide.OVER, RelaxationSide.BOTH}:
        plotly_data.append(go.Surface(x=x_list, y=y_list, z=w_max, name='overestimator'))

    fig = go.Figure(data=plotly_data)
    if show_plot:
        fig.show()

    x.unfix()
    y.unfix()
    x.value = orig_xval
    y.value = orig_yval
    w.value = orig_wval
    del m._underestimator_obj
    del m._overestimator_obj
    if orig_obj is not None:
        orig_obj.activate()


def plot_relaxation(m, relaxation, solver, show_plot=True, num_pts=100):
    rhs_vars = relaxation.get_rhs_vars()

    if len(rhs_vars) == 1:
        _plot_2d(m, relaxation, solver, show_plot, num_pts)
    elif len(rhs_vars) == 2:
        _plot_3d(m, relaxation, solver, show_plot, num_pts)
    else:
        raise NotImplementedError('Cannot generate plot for relaxation with more than 2 RHS vars')




from coramin.utils.coramin_enums import RelaxationSide, FunctionShape
from coramin.relaxations.custom_block import declare_custom_block
from coramin.relaxations.relaxations_base import BaseRelaxationData, ComponentWeakRef
from pyomo.core.expr.visitor import identify_variables
import math
import pyomo.environ as pe
from coramin.relaxations._utils import _get_bnds_list


@declare_custom_block(name='MultivariateRelaxation')
class MultivariateRelaxationData(BaseRelaxationData):
    def __init__(self, component):
        super(MultivariateRelaxationData, self).__init__(component)
        self._xs = None
        self._aux_var_ref = ComponentWeakRef(None)
        self._f_x_expr = None
        self._function_shape = FunctionShape.UNKNOWN

    @property
    def _aux_var(self):
        return self._aux_var_ref.get_component()

    def get_rhs_vars(self):
        return self._xs

    def get_rhs_expr(self):
        return self._f_x_expr

    def vars_with_bounds_in_relaxation(self):
        return list()

    def set_input(self, aux_var, shape, f_x_expr, use_linear_relaxation=True, large_coef=1e5, small_coef=1e-10,
                  safety_tol=1e-10):
        """
        Parameters
        ----------
        aux_var: pyomo.core.base.var._GeneralVarData
            The auxiliary variable replacing f(x)
        shape: FunctionShape
            Either FunctionShape.CONVEX or FunctionShape.CONCAVE
        f_x_expr: pyomo expression
            The pyomo expression representing f(x)
        use_linear_relaxation: bool
            Specifies whether a linear or nonlinear relaxation should be used
        """
        if shape not in {FunctionShape.CONVEX, FunctionShape.CONCAVE}:
            raise ValueError('MultivariateRelaxation only supports concave or convex functions.')
        self._function_shape = shape
        if shape == FunctionShape.CONVEX:
            relaxation_side = RelaxationSide.UNDER
        else:
            relaxation_side = RelaxationSide.OVER
        super().set_input(relaxation_side=relaxation_side,
                          use_linear_relaxation=use_linear_relaxation,
                          large_coef=large_coef, small_coef=small_coef,
                          safety_tol=safety_tol)
        self._xs = tuple(identify_variables(f_x_expr, include_fixed=False))
        self._aux_var_ref.set_component(aux_var)
        self._f_x_expr = f_x_expr

    def build(self, aux_var, shape, f_x_expr, use_linear_relaxation=True, large_coef=1e5, small_coef=1e-10,
              safety_tol=1e-10):
        self.set_input(aux_var=aux_var, shape=shape, f_x_expr=f_x_expr, use_linear_relaxation=use_linear_relaxation,
                       large_coef=large_coef, small_coef=small_coef, safety_tol=safety_tol)
        self.rebuild()

    def is_rhs_convex(self):
        return self._function_shape == FunctionShape.CONVEX

    def is_rhs_concave(self):
        return self._function_shape == FunctionShape.CONCAVE

    @property
    def use_linear_relaxation(self):
        return self._use_linear_relaxation

    @use_linear_relaxation.setter
    def use_linear_relaxation(self, value):
        self._use_linear_relaxation = value

    @property
    def relaxation_side(self):
        return BaseRelaxationData.relaxation_side.fget(self)

    @relaxation_side.setter
    def relaxation_side(self, val):
        if self.is_rhs_convex():
            if val != RelaxationSide.UNDER:
                raise ValueError('MultivariateRelaxations only support underestimators for convex functions')
        if self.is_rhs_concave():
            if val != RelaxationSide.OVER:
                raise ValueError('MultivariateRelaxations only support overestimators for concave functions')
        BaseRelaxationData.relaxation_side.fset(self, val)



from setuptools import setup, find_packages
from distutils.core import Extension


setup(name='coramin',
      version='0.2.2',
      packages=find_packages(),
      ext_modules=[],
      description='Coramin: Pyomo tools for MINLP',
      long_description=open('README.md').read(),
      long_description_content_type="text/markdown",
      author='Coramin Developers',
      maintainer_email='mlbynum@sandia.gov',
      license='Revised BSD',
      url='https://github.com/Coramin/Coramin',
      install_requires=['pyomo>=6.4.1', 'numpy', 'scipy', 'networkx'],
      include_package_data=True,
      scripts=[],
      python_requires='>=3.7',
      classifiers=["Programming Language :: Python :: 3",
                   "Programming Language :: Python :: 3.7",
                   "Programming Language :: Python :: 3.8",
                   "Programming Language :: Python :: 3.9",
                   "License :: OSI Approved :: BSD License",
                   "Operating System :: OS Independent"])



from mpi4py import MPI
import numpy as np
import sys
import os


class MPISyncError(Exception):
    pass


class MPIInterface:
    def __init__(self):
        self._comm = MPI.COMM_WORLD
        self._size = self._comm.Get_size()
        self._rank = self._comm.Get_rank()

    @property
    def comm(self):
        return self._comm
    
    @property
    def rank(self):
        return self._rank
    
    @property
    def size(self):
        return self._size


class MPIAllocationMap:
    def __init__(self, mpi_interface, global_N):
        self._mpi_interface = mpi_interface
        self._global_N = global_N

        rank = self._mpi_interface.rank
        size = self._mpi_interface.size

        # there must be a better way to do this
        # find which entries in global correspond
        # to this process (want them to be contiguous
        # for the MPI Allgather calls later
        local_N = [0 for i in range(self._mpi_interface.size)]
        for i in range(global_N):
            process_i = i % size
            local_N[process_i] += 1

        start = 0
        end = None
        for i,v in enumerate(local_N):
            if i == self._mpi_interface.rank:
                end = start + v
                break
            else:
                start += v

        self._local_map = list(range(start, end))

    def local_allocation_map(self):
        return list(self._local_map)
    
    def local_list(self, global_data):
        local_data = list()
        assert(len(global_data) == self._global_N)
        for i in self._local_map:
            local_data.append(global_data[i])
        return local_data

    def global_list_float64(self, local_data_float64):
        assert(len(local_data_float64) == len(self._local_map))
        global_data_numpy = np.zeros(self._global_N, dtype='d')*np.nan
        local_data_numpy = np.asarray(local_data_float64, dtype='d')
        comm = self._mpi_interface.comm
        comm.Allgatherv([local_data_numpy, MPI.DOUBLE],
                        [global_data_numpy, MPI.DOUBLE])

        return global_data_numpy.tolist()


def activate_mpi_printing(style='rank-0-console', rank_0_filename='output_rank_0.txt'):
    """
    Redirect standard output based on process rank.
    
    Parameters
    ----------
    style: str
        Can be set to one of:
           * 'ignore-all': ignore all printing (actually, redirect all printing to os.devnull)
           * 'rank-0-console': printing from rank 0 will go to the console, 
             printing from other processes will be ignored
           * 'rank-0-console-x-files': printing from rank 0 will go to the console, 
             printing from other processes will go to a separate file ('output_rank_x.txt')
           * 'rank-0-file': printing from rank 0 will go to 'output_rank_0.txt'
           * 'separate-files': printing from each processor will be redirected to a separate
             file for each process ('output_rank_x.txt')
    """
    rank = MPIInterface().rank
    if style == 'ignore-all':
        sys.stdout = open(os.devnull, 'w')
    elif style == 'rank-0-console':
        if rank != 0:
            sys.stdout = open(os.devnull, 'w')
    elif style == 'rank-0-file':
        if rank == 0:
            sys.stdout = open(rank_0_filename, 'w')
        else:
            sys.stdout = open(os.devnull, 'w')
    elif style == 'rank-0-console-x-files':
        if rank != 0:
            sys.stdout = open('output_rank_{0}.txt'.format(str(MPIInterface().rank)), 'w')
    elif style == 'separate-files':
        sys.stdout = open('output_rank_{0}.txt'.format(str(MPIInterface().rank)), 'w')



from pyomo.common.collections import ComponentSet
from coramin.domain_reduction.obbt import _bt_prep, _bt_cleanup
import pyomo.environ as pe
from pyomo.core.expr.current import LinearExpression
import logging
from pyomo.contrib import appsi
from pyomo.core.base.var import _GeneralVarData
from pyomo.core.base.block import _BlockData
from typing import Sequence, Optional, Union


logger = logging.getLogger(__name__)


def filter_variables_from_solution(candidate_variables_at_relaxation_solution, tolerance=1e-6):
    """
    This function takes a set of candidate variables for OBBT and filters out 
    the variables that are at their bounds in the provided solution to the 
    relaxation. See 

    Gleixner, Ambros M., et al. "Three enhancements for
    optimization-based bound tightening." Journal of Global
    Optimization 67.4 (2017): 731-757.

    for details on why this works. The basic idea is that if x = xl is
    feasible for the relaxation that will be used for OBBT, then
    minimizing x subject to that relaxation is guaranteed to result in
    an optimal solution of x* = xl.

    This function simply loops through
    candidate_variables_at_relaxation_solution and specifies which
    variables should be minimized and which variables should be
    maximized with OBBT.

    Parameters
    ----------
    candidate_variables_at_relaxation_solution: iterable of _GeneralVarData
        This should be an iterable of the variables which are candidates 
        for OBBT. The values of the variables should be feasible for the 
        relaxation that would be used to perform OBBT on the variables.
    tolerance: float
        A float greater than or equal to zero. If the value of the variable
        is within tolerance of its lower bound, then that variable is filtered
        from the set of variables that should be minimized for OBBT. The same
        is true for upper bounds and variables that should be maximized.

    Returns
    -------
    vars_to_minimize: ComponentSet of _GeneralVarData
        variables that should be considered for minimization
    vars_to_maximize: ComponentSet of _GeneralVarData
        variables that should be considered for maximization
    """
    candidate_vars = ComponentSet(candidate_variables_at_relaxation_solution)
    vars_to_minimize = ComponentSet()
    vars_to_maximize = ComponentSet()

    for v in candidate_vars:
        if (not v.has_lb()) or (v.value - v.lb > tolerance):
            vars_to_minimize.add(v)
        if (not v.has_ub()) or (v.ub - v.value > tolerance):
            vars_to_maximize.add(v)

    return vars_to_minimize, vars_to_maximize


def aggressive_filter(
    candidate_variables: Sequence[_GeneralVarData],
    relaxation: _BlockData,
    solver: Union[appsi.base.Solver, appsi.base.PersistentSolver],
    tolerance: float = 1e-6,
    objective_bound: Optional[float] = None,
    max_iter: int = 10,
    improvement_threshold: int = 5
):
    """
    This function takes a set of candidate variables for OBBT and filters out 
    the variables for which it does not make senese to perform OBBT on. See 

    Gleixner, Ambros M., et al. "Three enhancements for
    optimization-based bound tightening." Journal of Global
    Optimization 67.4 (2017): 731-757.

    for details. The basic idea is that if x = xl is
    feasible for the relaxation that will be used for OBBT, then
    minimizing x subject to that relaxation is guaranteed to result in
    an optimal solution of x* = xl.

    This function solves a series of optimization problems to try to 
    filter as many variables as possible.

    Parameters
    ----------
    candidate_variables: iterable of _GeneralVarData
        This should be an iterable of the variables which are candidates 
        for OBBT.
    relaxation: Block
        a convex relaxation
    solver: appsi.base.Solver
    tolerance: float
        A float greater than or equal to zero. If the value of the variable
        is within tolerance of its lower bound, then that variable is filtered
        from the set of variables that should be minimized for OBBT. The same
        is true for upper bounds and variables that should be maximized.
    objective_bound: float
        Primal bound for the objective
    max_iter: int
        Maximum number of iterations
    improvement_threshold: int
        If the number of filtered variables is less than improvement_threshold, then
        the filtering is terminated

    Returns
    -------
    vars_to_minimize: list of _GeneralVarData
        variables that should be considered for minimization
    vars_to_maximize: list of _GeneralVarData
        variables that should be considered for maximization
    """
    vars_to_minimize = ComponentSet(candidate_variables)
    vars_to_maximize = ComponentSet(candidate_variables)
    if len(candidate_variables) == 0:
        return vars_to_minimize, vars_to_maximize

    tmp = _bt_prep(model=relaxation, solver=solver, objective_bound=objective_bound)
    initial_var_values, deactivated_objectives, orig_update_config, orig_config = tmp

    vars_unbounded_from_below = ComponentSet()
    vars_unbounded_from_above = ComponentSet()
    for v in list(vars_to_minimize):
        if v.lb is None:
            vars_unbounded_from_below.add(v)
            vars_to_minimize.remove(v)
    for v in list(vars_to_maximize):
        if v.ub is None:
            vars_unbounded_from_above.add(v)
            vars_to_maximize.remove(v)

    for _set in [vars_to_minimize, vars_to_maximize]:
        for _iter in range(max_iter):
            if _set is vars_to_minimize:
                obj_coefs = [1 for v in _set]
            else:
                obj_coefs = [-1 for v in _set]
            obj_vars = list(_set)
            relaxation.__filter_obj = pe.Objective(expr=LinearExpression(linear_coefs=obj_coefs, linear_vars=obj_vars))
            if solver.is_persistent():
                solver.set_objective(relaxation.__filter_obj)
            solver.config.load_solution = False
            res = solver.solve(relaxation)
            if res.termination_condition == appsi.base.TerminationCondition.optimal:
                res.solution_loader.load_vars()
                success = True
            else:
                success = False
            del relaxation.__filter_obj

            if not success:
                break

            num_filtered = 0
            for v in list(_set):
                should_filter = False
                if _set is vars_to_minimize:
                    if v.value - v.lb <= tolerance:
                        should_filter = True
                else:
                    if v.ub - v.value <= tolerance:
                        should_filter = True
                if should_filter:
                    num_filtered += 1
                    _set.remove(v)
            logger.debug('filtered {0} vars on iter {1}'.format(num_filtered, _iter))

            if len(_set) == 0:
                break
            if num_filtered < improvement_threshold:
                break

    for v in vars_unbounded_from_below:
        vars_to_minimize.add(v)
    for v in vars_unbounded_from_above:
        vars_to_maximize.add(v)

    _bt_cleanup(
        model=relaxation, solver=solver, vardatalist=None,
        initial_var_values=initial_var_values,
        deactivated_objectives=deactivated_objectives,
        orig_update_config=orig_update_config, orig_config=orig_config,
        lower_bounds=None, upper_bounds=None
    )

    return vars_to_minimize, vars_to_maximize



import pyomo.environ as pyo
from pyomo.core.base.var import _GeneralVarData
from pyomo.core.base.PyomoModel import ConcreteModel
import warnings
from pyomo.common.collections import ComponentMap
from pyomo.core.kernel.objective import minimize, maximize
from pyomo.contrib import appsi
import logging
import traceback
import numpy as np
import math
import time
from typing import Union, Sequence, Optional, List
from pyomo.core.base.objective import ScalarObjective, _GeneralObjectiveData
try:
    import coramin.utils.mpi_utils as mpiu
    mpi_available = True
except ImportError:
    mpi_available = False
try:
    from tqdm import tqdm
except ImportError:
    pass


logger = logging.getLogger(__name__)


class OBBTInfo(object):
    def __init__(self):
        self.total_num_problems = None
        self.num_problems_attempted = None
        self.num_successful_problems = None


def _bt_cleanup(
    model, solver: Union[appsi.base.Solver, appsi.base.PersistentSolver],
    vardatalist: Optional[List[_GeneralVarData]],
    initial_var_values, deactivated_objectives, orig_update_config, orig_config,
    lower_bounds: Optional[Sequence[float]] = None,
    upper_bounds: Optional[Sequence[float]] = None
):
    """
    Cleanup the changes made to the model during bounds tightening.
    Reactivate any deactivated objectives.
    Remove an objective upper bound constraint if it was added.
    If lower_bounds or upper_bounds is provided, update the bounds of the variables in self.vars_to_tighten.

    Parameters
    ----------
    model: pyo.ConcreteModel or pyo.Block
    solver: Union[appsi.base.Solver, appsi.base.PersistentSolver]
    vardatalist: List of _GeneralVarData
    initial_var_values: ComponentMap
    deactivated_objectives: list of _GeneralObjectiveData
    orig_update_config: appsi.base.UpdateConfig
    orig_config: appsi.base.SolverConfig
    lower_bounds: Sequence of float
        Only needed if you want to update the bounds of the variables. Should be in the same order as
        self.vars_to_tighten.
    upper_bounds: Sequence of float
        Only needed if you want to update the bounds of the variables. Should be in the same order as
        self.vars_to_tighten.
    """
    for v in model.component_data_objects(ctype=pyo.Var, active=None, sort=True, descend_into=True):
        v.set_value(initial_var_values[v], skip_validation=True)

    if hasattr(model, '__objective_ineq'):
        if solver.is_persistent():
            solver.remove_constraints([model.__objective_ineq])
        del model.__objective_ineq

    # reactivate the objectives that we deactivated
    for obj in deactivated_objectives:
        obj.activate()
        if solver.is_persistent():
            solver.set_objective(obj)

    if lower_bounds is not None and upper_bounds is not None:
        for i, v in enumerate(vardatalist):
            lb = lower_bounds[i]
            ub = upper_bounds[i]
            v.setlb(lb)
            v.setub(ub)
    elif lower_bounds is not None:
        for i, v in enumerate(vardatalist):
            lb = lower_bounds[i]
            v.setlb(lb)
    elif upper_bounds is not None:
        for i, v in enumerate(vardatalist):
            ub = upper_bounds[i]
            v.setub(ub)
    if vardatalist is not None and solver.is_persistent():
        solver.update_variables(vardatalist)

    if solver.is_persistent():
        solver.update_config.check_for_new_or_removed_constraints = \
            orig_update_config.check_for_new_or_removed_constraints
        solver.update_config.check_for_new_or_removed_vars = \
            orig_update_config.check_for_new_or_removed_vars
        solver.update_config.check_for_new_or_removed_params = \
            orig_update_config.check_for_new_or_removed_params
        solver.update_config.check_for_new_objective = \
            orig_update_config.check_for_new_objective
        solver.update_config.update_constraints = \
            orig_update_config.update_constraints
        solver.update_config.update_vars = \
            orig_update_config.update_vars
        solver.update_config.update_params = \
            orig_update_config.update_params
        solver.update_config.update_named_expressions = \
            orig_update_config.update_named_expressions
        solver.update_config.update_objective = \
            orig_update_config.update_objective
        solver.update_config.treat_fixed_vars_as_params = \
            orig_update_config.treat_fixed_vars_as_params

    solver.config.stream_solver = orig_config.stream_solver
    solver.config.load_solution = orig_config.load_solution


def _single_solve(v, model, solver: Union[appsi.base.Solver, appsi.base.PersistentSolver], lb_or_ub, obbt_info):
    obbt_info.num_problems_attempted += 1
    # solve for lower var bound
    if lb_or_ub == 'lb':
        model.__obj_bounds_tightening = ScalarObjective(expr=v, sense=pyo.minimize)
    else:
        assert lb_or_ub == 'ub'
        model.__obj_bounds_tightening = ScalarObjective(expr=v, sense=pyo.maximize)

    if solver.is_persistent():
        solver.set_objective(model.__obj_bounds_tightening)
    results = solver.solve(model)
    if results.termination_condition == appsi.base.TerminationCondition.optimal:
        obbt_info.num_successful_problems += 1
    if results.best_objective_bound is not None and math.isfinite(results.best_objective_bound):
        new_bnd = results.best_objective_bound
    elif results.termination_condition == appsi.base.TerminationCondition.optimal:
        new_bnd = results.best_feasible_objective  # assumes the problem is convex
    else:
        new_bnd = None
        msg = f'Warning: Bounds tightening for lb for var {str(v)} was unsuccessful. Termination condition: {results.termination_condition}; The lb was not changed.'
        logger.warning(msg)

    if lb_or_ub == 'lb':
        orig_lb = pyo.value(v.lb)
        if new_bnd is None:
            new_bnd = orig_lb
        elif v.has_lb():
            if new_bnd < orig_lb:
                new_bnd = orig_lb
    else:
        orig_ub = pyo.value(v.ub)
        if new_bnd is None:
            new_bnd = orig_ub
        elif v.has_ub():
            if new_bnd > orig_ub:
                new_bnd = orig_ub

    if new_bnd is None:
        # Need nan instead of None for MPI communication; This is appropriately handled in perform_obbt().
        new_bnd = np.nan

    # remove the objective function
    del model.__obj_bounds_tightening

    return new_bnd


def _tighten_bnds(model, solver, vardatalist, lb_or_ub, obbt_info, with_progress_bar=False, time_limit=math.inf, progress_bar_string=None):
    """
    Tighten the lower bounds of all variables in vardatalist (or self.vars_to_tighten if vardatalist is None).

    Parameters
    ----------
    model: pyo.ConcreteModel or pyo.Block
    solver: pyomo solver object
    vardatalist: list of _GeneralVarData
    lb_or_ub: str
        'lb' or 'ub'
    time_limit: float

    Returns
    -------
    new_bounds: list of float
    """
    # solve for the new bounds
    t0 = time.time()
    new_bounds = list()

    obbt_info.total_num_problems += len(vardatalist)

    if with_progress_bar:
        if progress_bar_string is None:
            if lb_or_ub == 'lb':
                bnd_str = 'LBs'
            else:
                bnd_str = 'UBs'
            bnd_str = 'OBBT ' + bnd_str
        else:
            bnd_str = progress_bar_string
        if mpi_available:
            tqdm_position = mpiu.MPI.COMM_WORLD.Get_rank()
        else:
            tqdm_position = 0
        for v in tqdm(vardatalist, ncols=100, desc=bnd_str, leave=False, position=tqdm_position):
            if time.time() - t0 > time_limit:
                if lb_or_ub == 'lb':
                    if v.lb is None:
                        new_bounds.append(np.nan)
                    else:
                        new_bounds.append(pyo.value(v.lb))
                else:
                    if v.ub is None:
                        new_bounds.append(np.nan)
                    else:
                        new_bounds.append(pyo.value(v.ub))
            else:
                new_bnd = _single_solve(v=v, model=model, solver=solver,
                                        lb_or_ub=lb_or_ub,
                                        obbt_info=obbt_info)
                new_bounds.append(new_bnd)
    else:
        for v in vardatalist:
            if time.time() - t0 > time_limit:
                if lb_or_ub == 'lb':
                    if v.lb is None:
                        new_bounds.append(np.nan)
                    else:
                        new_bounds.append(pyo.value(v.lb))
                else:
                    if v.ub is None:
                        new_bounds.append(np.nan)
                    else:
                        new_bounds.append(pyo.value(v.ub))
            else:
                new_bnd = _single_solve(v=v, model=model, solver=solver,
                                        lb_or_ub=lb_or_ub,
                                        obbt_info=obbt_info)
                new_bounds.append(new_bnd)

    return new_bounds


def _bt_prep(model, solver, objective_bound=None):
    """
    Prepare the model for bounds tightening.
    Gather the variable values to load back in after bounds tightening.
    Deactivate any active objectives.
    If objective_ub is not None, then add a constraint forcing the objective to be less than objective_ub

    Parameters
    ----------
    model : pyo.ConcreteModel or pyo.Block
        The model object that will be used for bounds tightening.
    objective_bound : float
        The objective value for the current best upper bound incumbent

    Returns
    -------
    initial_var_values: ComponentMap
    deactivated_objectives: list
    orig_update_config: appsi.base.UpdateConfig
    orig_config: appsi.base.SolverConfig
    """

    if solver.is_persistent():
        orig_update_config = solver.update_config()
        solver.update_config.check_for_new_or_removed_constraints = False
        solver.update_config.check_for_new_or_removed_vars = False
        solver.update_config.check_for_new_or_removed_params = False
        solver.update_config.check_for_new_objective = False
        solver.update_config.update_constraints = False
        solver.update_config.update_vars = False
        solver.update_config.update_params = False
        solver.update_config.update_named_expressions = False
        solver.update_config.update_objective = False
        solver.update_config.treat_fixed_vars_as_params = True
    else:
        orig_update_config = None

    orig_config = solver.config()
    solver.config.stream_solver = False
    solver.config.load_solution = False

    if solver.is_persistent():
        solver.set_instance(model)

    initial_var_values = ComponentMap()
    for v in model.component_data_objects(ctype=pyo.Var, active=None, sort=True, descend_into=True):
        initial_var_values[v] = v.value

    deactivated_objectives = list()
    for obj in model.component_data_objects(pyo.Objective, active=True, sort=True, descend_into=True):
        deactivated_objectives.append(obj)
        obj.deactivate()

    # add inequality bound on objective functions if required
    # obj.expr <= objective_ub
    if objective_bound is not None and math.isfinite(objective_bound):
        if len(deactivated_objectives) != 1:
            e = 'BoundsTightener: When providing objective_ub,' + \
                ' the model must have one and only one objective function.'
            logger.error(e)
            raise ValueError(e)
        original_obj = deactivated_objectives[0]
        if original_obj.sense == minimize:
            model.__objective_ineq = \
                pyo.Constraint(expr=original_obj.expr <= objective_bound)
        else:
            assert original_obj.sense == maximize
            model.__objective_ineq = pyo.Constraint(expr=original_obj.expr >= objective_bound)
        if solver.is_persistent():
            solver.add_constraints([model.__objective_ineq])

    return initial_var_values, deactivated_objectives, orig_update_config, orig_config


def _build_vardatalist(model, varlist=None, warning_threshold=0):
    """
    Convert a list of pyomo variables to a list of SimpleVar and _GeneralVarData. If varlist is none, builds a
    list of all variables in the model. The new list is stored in the vars_to_tighten attribute.

    Parameters
    ----------
    model: ConcreteModel
    varlist: None or list of pyo.Var
    warning_threshold: float
        The threshold below which a warning is raised when attempting to perform OBBT on variables whose
        ub - lb < warning_threshold.
    """
    vardatalist = None

    # if the varlist is None, then assume we want all the active variables
    if varlist is None:
        raise NotImplementedError('Still need to do this.')
    elif isinstance(varlist, pyo.Var):
        # user provided a variable, not a list of variables. Let's work with it anyway
        varlist = [varlist]

    if vardatalist is None:
        # expand any indexed components in the list to their
        # component data objects
        vardatalist = list()
        for v in varlist:
            if v.is_indexed():
                vardatalist.extend(v.values())
            else:
                vardatalist.append(v)

    # remove from vardatalist if the variable is fixed (maybe there is a better way to do this)
    corrected_vardatalist = []
    for v in vardatalist:
        if not v.is_fixed():
            if v.has_lb() and v.has_ub():
                if v.ub - v.lb < warning_threshold:
                    e = 'Warning: Tightening a variable with ub - lb is less than {threshold}: {v}, lb: {lb}, ub: {ub}'.format(threshold=warning_threshold, v=v, lb=v.lb, ub=v.ub)
                    logger.warning(e)
                    warnings.warn(e)
            corrected_vardatalist.append(v)

    return corrected_vardatalist


def perform_obbt(model, solver, varlist=None, objective_bound=None, update_bounds=True, with_progress_bar=False,
                 direction='both', time_limit=math.inf, parallel=True, collect_obbt_info=False,
                 warning_threshold=0, progress_bar_string=None):
    """
    Perform optimization-based bounds tighening on the variables in varlist subject to the constraints in model.

    Parameters
    ----------
    model: pyo.ConcreteModel or pyo.Block
        The model to be used for bounds tightening
    solver: appsi.base.PersistentSolver
        The solver to be used for bounds tightening.
    varlist: list of pyo.Var
        The variables for which OBBT should be performed. If varlist is None, then we attempt to automatically
        detect which variables need tightened.
    objective_bound: float
        A lower or upper bound on the objective. If this is not None, then a constraint will be added to the
        bounds tightening problems constraining the objective to be less than/greater than objective_bound.
    update_bounds: bool
        If True, then the variable bounds will be updated
    with_progress_bar: bool
    direction: str
        Options are 'both', 'lbs', or 'ubs'
    time_limit: float
        The maximum amount of time to be spent performing OBBT
    parallel: bool
        If True, then OBBT will automatically be performed in parallel if mpirun or mpiexec was used;
        If False, then OBBT will not run in parallel even if mpirun or mpiexec was used;
    warning_threshold: float
        The threshold below which a warning is issued when attempting to perform OBBT on variables whose
        ub - lb < warning_threshold.

    Returns
    -------
    lower_bounds: list of float
    upper_bounds: list of float
    obbt_info: OBBTInfo

    """
    if not isinstance(solver, appsi.base.Solver):
        raise ValueError('Coramin requires an Appsi solver interface')

    obbt_info = OBBTInfo()
    obbt_info.total_num_problems = 0
    obbt_info.num_problems_attempted = 0
    obbt_info.num_successful_problems = 0

    t0 = time.time()
    initial_var_values, deactivated_objectives, orig_update_config, orig_config = _bt_prep(model=model, solver=solver, objective_bound=objective_bound)

    vardata_list = _build_vardatalist(model=model, varlist=varlist, warning_threshold=warning_threshold)
    if mpi_available and parallel:
        mpi_interface = mpiu.MPIInterface()
        alloc_map = mpiu.MPIAllocationMap(mpi_interface, len(vardata_list))
        local_vardata_list = alloc_map.local_list(vardata_list)
    else:
        local_vardata_list = vardata_list

    exc = None
    try:
        if direction in {'both', 'lbs'}:
            local_lower_bounds = _tighten_bnds(model=model, solver=solver,
                                               vardatalist=local_vardata_list,
                                               lb_or_ub='lb',
                                               obbt_info=obbt_info,
                                               with_progress_bar=with_progress_bar,
                                               time_limit=(time_limit - (time.time() - t0)),
                                               progress_bar_string=progress_bar_string)
        else:
            local_lower_bounds = list()
            for v in local_vardata_list:
                if v.lb is None:
                    local_lower_bounds.append(np.nan)
                else:
                    local_lower_bounds.append(pyo.value(v.lb))
        if direction in {'both', 'ubs'}:
            local_upper_bounds = _tighten_bnds(model=model, solver=solver,
                                               vardatalist=local_vardata_list,
                                               lb_or_ub='ub',
                                               obbt_info=obbt_info,
                                               with_progress_bar=with_progress_bar,
                                               time_limit=(time_limit - (time.time() - t0)),
                                               progress_bar_string=progress_bar_string)
        else:
            local_upper_bounds = list()
            for v in local_vardata_list:
                if v.ub is None:
                    local_upper_bounds.append(np.nan)
                else:
                    local_upper_bounds.append(pyo.value(v.ub))
        status = 1
        msg = None
    except Exception as err:
        exc = err
        tb = traceback.format_exc()
        status = 0
        msg = str(tb)

    if mpi_available and parallel:
        local_status = np.array([status], dtype='i')
        global_status = np.array([0 for i in range(mpiu.MPI.COMM_WORLD.Get_size())], dtype='i')
        mpiu.MPI.COMM_WORLD.Allgatherv([local_status, mpiu.MPI.INT], [global_status, mpiu.MPI.INT])
        if not np.all(global_status):
            messages = mpi_interface.comm.allgather(msg)
            msg = None
            for m in messages:
                if m is not None:
                    msg = m
            logger.error('An error was raised in one or more processes:\n' + msg)
            raise mpiu.MPISyncError('An error was raised in one or more processes:\n' + msg)
    else:
        if status != 1:
            logger.error('An error was raised during OBBT:\n' + msg)
            raise exc

    if mpi_available and parallel:
        global_lower = alloc_map.global_list_float64(local_lower_bounds)
        global_upper = alloc_map.global_list_float64(local_upper_bounds)
        obbt_info.total_num_problems = mpiu.MPI.COMM_WORLD.allreduce(obbt_info.total_num_problems)
        obbt_info.num_problems_attempted = mpiu.MPI.COMM_WORLD.allreduce(obbt_info.num_problems_attempted)
        obbt_info.num_successful_problems = mpiu.MPI.COMM_WORLD.allreduce(obbt_info.num_successful_problems)
    else:
        global_lower = local_lower_bounds
        global_upper = local_upper_bounds

    tmp = list()
    for i in global_lower:
        if np.isnan(i):
            tmp.append(None)
        else:
            tmp.append(float(i))
    global_lower = tmp

    tmp = list()
    for i in global_upper:
        if np.isnan(i):
            tmp.append(None)
        else:
            tmp.append(float(i))
    global_upper = tmp

    _lower_bounds = None
    _upper_bounds = None
    if update_bounds:
        _lower_bounds = global_lower
        _upper_bounds = global_upper
    _bt_cleanup(
        model=model, solver=solver, vardatalist=vardata_list,
        initial_var_values=initial_var_values,
        deactivated_objectives=deactivated_objectives,
        orig_update_config=orig_update_config, orig_config=orig_config,
        lower_bounds=_lower_bounds, upper_bounds=_upper_bounds
    )

    if collect_obbt_info:
        return global_lower, global_upper, obbt_info
    else:
        return global_lower, global_upper
